/*
<one line to give the program's name and a brief idea of what it does.>
Copyright (C) <year>  <name of author>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "cpppropertycreator.h"
#include "propertynode.h"
#include "propertymodel.h"

#include <language/duchain/duchainlock.h>
#include <language/duchain/duchain.h>
#include <language/duchain/classdeclaration.h>
#include <language/duchain/declaration.h>
#include <language/duchain/duchainutils.h>
#include <language/duchain/classfunctiondeclaration.h>
#include <language/duchain/types/functiontype.h>
#include <interfaces/idocument.h>
#include <interfaces/icore.h>
#include <interfaces/idocumentcontroller.h>

#include <KTextEditor/View>
#include <KTextEditor/Document>
#include <KMessageBox>
#include <interfaces/isourceformattercontroller.h>
#include <interfaces/isourceformatter.h>



CppPropertyCreator::CodeRange& CppPropertyCreator::getRange (const QString& name)
{
    static NullCodeRange r;
    for (int i = 0; i < m_ranges.size(); i++)
    {
        if (m_ranges[i].name == name)
            return m_ranges[i];
    }
    return r;
}


int CppPropertyCreator::getRangeIndex (const QString& name)
{
    for (int i = 0; i < m_ranges.size(); i++)
    {
        if (m_ranges[i].name == name)
            return i;
    }
    return -1;
}


void CppPropertyCreator::create (PropertyModel* m)
{
    DUChainReadLocker lock (DUChain::lock());

    ClassDeclaration * cd = m->klass();

    IDocument* doc = ICore::self()->documentController()->documentForUrl (cd->topContext()->url().toUrl());
    if (!doc)
    {
        kDebug() << "No document";
        return;
    }

    KTextEditor::Document* textDoc = doc->textDocument();
    if (!textDoc)
    {
        kDebug() << "No text document";
        return;
    }

    KTextEditor::View* view = textDoc->activeView();
    if (!view)
    {
        kDebug() << "No view";
        return;
    }

    QString text = textDoc->text (m->klass()->internalContext()->range().textRange());

    int lineCount = m->klass()->internalContext()->range().start.line;
    int dummy = 0, *end = &dummy;

    QRegExp publicExp ("^\\s*public\\s*:");
    QRegExp protectedExp ("^\\s*protected\\s*:");
    QRegExp privateExp ("^\\s*private\\s*:");
    QRegExp indentExp ("^(\\s+)(\\w+.*)");
    QStringList lines = text.split ('\n');
    QString indent = "";
    m_ranges.clear();

    foreach (QString line, lines)
    {
        //Skip inline classes
        DUContext *duc = m->klass()->internalContext()->findContextAt (SimpleCursor (lineCount, 0));
        if (duc && duc != m->klass()->internalContext())
        {
            lineCount++;
            continue;
        }

        if (line.contains (publicExp))
        {
            if (getRange ("public").isNull())
            {
                m_ranges.append (CodeRange ("public"));
            }
            getRange ("public").lineStart = lineCount;
            *end = lineCount;
            end = & (getRange ("public").lineEnd);
        }
        else if (line.contains (protectedExp))
        {
            if (getRange ("protected").isNull())
            {
                m_ranges.append (CodeRange ("protected"));
            }
            getRange ("protected").lineStart = lineCount;
            *end = lineCount;
            end = & (getRange ("protected").lineEnd);
        }
        else if (line.contains (privateExp))
        {
            if (getRange ("private").isNull())
            {
                m_ranges.append (CodeRange ("private"));
            }
            getRange ("private").lineStart = lineCount;
            *end = lineCount;
            end = & (getRange ("private").lineEnd);
        }
        else if (indentExp.exactMatch (line))
        {
            indent = indentExp.cap (1);
        }

        lineCount++;
    }

    if (*end == 0)
        *end = m->klass()->internalContext()->range().end.line;

    foreach (PropertyNode n, m->nodes())
    {
        if (n.createReader() && n.canCreateReader())
        {
            QStringList itext;
            QString text = QString ("%1 %2() const { return %3; }").arg (readerType (n.type()), n.readerName(), n.name());

            switch (m->modifier())
            {
            case PropertyModel::VIRTUAL:
                text = QString ("virtual ") + text;
                break;
            case PropertyModel::PURE_VIRTUAL:
                text = QString ("virtual %1 %2() const = 0;").arg (readerType (n.type()), n.readerName());
                break;
            case PropertyModel::INLINE:
                text = QString ("inline ") + text;
                break;
            }

            CodeRange &range = getRange (n.readerAccess());
            if (range.isNull())
            {
                m_ranges.append (CodeRange (n.readerAccess(), lineCount - 1, lineCount));
                textDoc->insertLine (lineCount - 1, QString ("%1:").arg (n.readerAccess()));
                lineCount += 1;
                range = getRange (n.readerAccess());
            }

            if (m->createDocumentation())
            {
                itext.append (QString ("/** Reader for %1.").arg (n.name()));
                itext.append (QString (" *  @see %1").arg (n.name()));
                itext.append (QString (" *  @return %1").arg (n.name()));
                itext.append (" */");
            }
            itext.append (text);

            if (ICore::self() && doc)
            {
                text = ICore::self()->sourceFormatterController()->formatterForUrl (doc->url())->formatSource (itext.join ("\n") ,doc->mimeType());
                itext = text.split ("\n");
            }

            for (int i= 0; i < itext.size(); i++)
            {
                textDoc->insertLine (range.lineEnd, indent + itext[i]);
                range.lineEnd++;
                lineCount++;
            }
            for (int i = getRangeIndex (n.readerAccess()) + 1; i < m_ranges.size(); i++)
            {
                m_ranges[i].lineStart += itext.size();
                m_ranges[i].lineEnd += itext.size();
            }
        }

        if (n.createWriter() && n.canCreateWriter())
        {
            QStringList itext;
            QString text = QString ("void %1( %2 %4 ) { %3 = %4; }").arg (n.writerName(), writerType (n.type()), n.name(), propertyAlias (n.name()));

            switch (m->modifier())
            {
            case PropertyModel::VIRTUAL:
                text = QString ("virtual ") + text;
                break;
            case PropertyModel::PURE_VIRTUAL:
                text = QString ("virtual void %1( %2 %3 ) = 0;").arg (n.writerName(), writerType (n.type()), propertyAlias (n.name()));
                break;
            case PropertyModel::INLINE:
                text = QString ("inline ") + text;
                break;
            }

            CodeRange &range = getRange (n.writerAccess());
            if (range.isNull())
            {
                m_ranges.append (CodeRange (n.writerAccess(), lineCount - 1, lineCount));
                textDoc->insertLine (lineCount - 1, QString ("%1:").arg (n.writerAccess()));
                lineCount += 1;
                range = getRange (n.writerAccess());
            }

            if (m->createDocumentation())
            {
                itext.append (QString ("/** Writer for %1.").arg (n.name()));
                itext.append (QString (" *  @see %1").arg (n.name()));
                itext.append (QString (" *  @param %1 Value for %2.").arg (propertyAlias (n.name()), n.name()));
                itext.append (" */");

            }

            itext.append (text);

            if (ICore::self() && doc)
            {
                text = ICore::self()->sourceFormatterController()->formatterForUrl (doc->url())->formatSource (itext.join ("\n") ,doc->mimeType());
                itext = text.split ("\n");
            }

            for (int i= 0; i < itext.size(); i++)
            {
                textDoc->insertLine (range.lineEnd, indent + itext[i]);
                range.lineEnd++;
                lineCount++;
            }
            for (int i = getRangeIndex (n.readerAccess()) + 1; i < m_ranges.size(); i++)
            {
                m_ranges[i].lineStart += itext.size();
                m_ranges[i].lineEnd += itext.size();
            }
        }
    }
}


bool CppPropertyCreator::parse (PropertyModel* m, ClassDeclaration* klass, ClassMemberDeclaration* selectedMember)
{
    DUChainReadLocker lock (DUChain::lock());
    QList<PropertyNode> nodes;

    if (!klass)
        return false;

    m->setKlass (klass);
    m->setSelectedMember (selectedMember);

    foreach (Declaration* decl, klass->internalContext()->localDeclarations())
    {
        ClassMemberDeclaration* memDecl = dynamic_cast<ClassMemberDeclaration*> (decl);
        if (!memDecl)
            continue;

        if (! (!memDecl->isFunctionDeclaration() && memDecl->kind() == Declaration::Instance))
            continue;

        //TODO This depends on staticMetaObject being at the begining of the loop
        if (memDecl->identifier().toString() == "staticMetaObject")
        {
            m->setHasQObject (true);
            continue;
        }

        PropertyNode node;
        node.setIcon (DUChainUtils::iconForDeclaration (memDecl));
        node.setType (memDecl->abstractType()->toString());
        node.setName (memDecl->identifier().toString());
        node.setReaderName (readerName (node.name()));
        node.setWriterName (writerName (node.name()));

        if (memDecl == selectedMember)
        {
            node.setCreateReader (true);
            node.setCreateWriter (true);
        }

        foreach (Declaration * d, klass->internalContext()->localDeclarations())
        {
            ClassFunctionDeclaration * dm =  dynamic_cast<ClassFunctionDeclaration*> (d);
            if (!dm)
                continue;

            FunctionType::Ptr ft = dm->abstractType().cast<FunctionType>();
            if (!ft)
                continue;
            if (!ft->returnType())
                continue;

            QString access = "default";
            switch (dm->accessPolicy())
            {
            case Declaration::Public:
                access = "public";
                break;
            case Declaration::Protected:
                access = "protected";
                break;
            case Declaration::Private:
                access = "private";
                break;
            }

            if (dm->identifier().toString() == node.readerName() &&
                    ft->arguments().size() == 0 &&
                    ft->modifiers() & AbstractType::ConstModifier)
            {
                node.setCanCreateReader (false);
                node.setCreateReader (true);
                node.setReaderAccess (access);
                continue;
            }

            if (dm->identifier().toString() == node.writerName() &&
                    ft->returnType()->toString() == "void" &&
                    ft->arguments().size() == 1 &&
                    ft->arguments() [0]->toString() == writerType (node.type()))
            {
                node.setCanCreateWriter (false);
                node.setCreateWriter (true);
                node.setWriterAccess (access);
                continue;
            }
        }
        nodes.append (node);
    }

    m->setNodes (nodes);

    return true;
}


QString CppPropertyCreator::propertyAlias (const QString& property)
{
    QString p = propertyName (property);
    QString r = "";
    r += p[0];
    bool kase = false;
    for (int i = 1; i < p.size(); i++)
    {
        if (p[i].isUpper() && kase == false)
        {
            r += p[i].toLower();
            kase = true;
        }
        if (p[i].isLower() && kase == true)
            kase = false;
        if (p[i] == '_' && i < p.size() - 1)
            r += p[i+1];
    }
    return r;
}


QString CppPropertyCreator::propertyName (const QString& property)
{
    static QRegExp exp ("(^m_)|(^m)|(^[_]*)");
    QString n = property;
    n = n.replace (exp, "");
    n[0] = n[0].toLower();
    return n;
}

QString CppPropertyCreator::readerName (const QString& property)
{
    QString n = propertyName (property);
    if (n == property)
    {
        n[0] = n[0].toUpper();
        n = QString ("set%1").arg (n);
    }
    return n;
}

QString CppPropertyCreator::writerName (const QString& property)
{
    QString n = propertyName (property);
    n[0] = n[0].toUpper();
    return QString ("get%1").arg (n);
}

QString CppPropertyCreator::readerType (const QString& type)
{
    static QString primitives ("bool long long int short int ushort uint double float char uchar");
    if (type.contains ('*'))
        return  QString ("const %1").arg (type);
    if (primitives.contains (type))
        return type;
    return QString ("const %1&").arg (type);
}

QString CppPropertyCreator::writerType (const QString& type)
{
    static QString primitives ("bool long long int short int ushort uint double float char uchar");
    if (type.contains ('*'))
        return  QString ("%1").arg (type);
    if (primitives.contains (type))
        return type;
    return QString ("const %1&").arg (type);
}



// kate: indent-mode cstyle; space-indent on; indent-width 4; replace-tabs on;  replace-tabs on;  replace-tabs on;
