/***************************************************************************
 *   Copyright 2001 Bernd Gehrmann <bernd@kdevelop.org>                    *
 *   Copyright 2004-2005 Sascha Cunz <sascha@kdevelop.org>                 *
 *   Copyright 2005 Ian Reinhart Geiser <ian@geiseri.com>                  *
 *   Copyright 2007 Alexander Dymo <adymo@kdevelop.org>                    *
 *   Copyright 2008 Evgeniy Ivanov <powerfox@kde.ru>                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "readerwriterplugin.h"

#include <interfaces/icore.h>
#include <interfaces/contextmenuextension.h>
#include <interfaces/context.h>
#include <interfaces/idocument.h>
#include <interfaces/iuicontroller.h>
#include <interfaces/idocumentcontroller.h>

#include <language/duchain/declaration.h>
#include <language/duchain/duchainutils.h>
#include <language/duchain/duchain.h>
#include <language/duchain/duchainlock.h>
#include <language/duchain/classdeclaration.h>
#include <language/duchain/classmemberdeclaration.h>
#include <shell/sourceformattercontroller.h>
#include <interfaces/isourceformatter.h>

#include <KPluginFactory>
#include <KPluginLoader>
#include <KLocalizedString>
#include <KAboutData>
#include <KTextEditor/Document>
#include <KDebug>
#include <KMessageBox>
#include <KActionCollection>
#include <KAction>

#include <QtCore/QPair>
#include <QtCore/QList>

#include "propertiesdialog.h"
#include "propertymodel.h"
#include <ktexteditor/smartcursor.h>
#include "cpppropertycreator.h"




K_PLUGIN_FACTORY (KdevPlugFactory, registerPlugin<ReaderWriterPlugin>();)
K_EXPORT_PLUGIN (KdevPlugFactory (KAboutData ("ReaderWriter","ReaderWriter", ki18n ("Generate Reader Writer Properties"), "0.1", ki18n ("Generate Reader Writer Properties"), KAboutData::License_GPL)))

ReaderWriterPlugin::ReaderWriterPlugin (QObject *parent, const QVariantList &)
        :IPlugin (KdevPlugFactory::componentData(), parent)
{
    setXMLFile ("readerwriter.rc");

    m_action = actionCollection()->addAction ("generate_readerwriter");
    m_action->setIcon (KIcon ("kdevelop"));
    m_action->setText (i18n ("Generate properties.."));
    m_action->setShortcut (KShortcut ("Ctrl+Alt+p"));
    connect (m_action, SIGNAL (triggered (bool)), this, SLOT (slotCreateProperties()));
    m_action->setToolTip (i18n ("Generate Reader Writer Properties"));
    m_action->setWhatsThis (i18n ("<b>ReaderWriter</b><p> Generate Reader Writer Properties"));
    m_action->setEnabled (false);
}

ReaderWriterPlugin::~ReaderWriterPlugin()
{
}

void ReaderWriterPlugin::slotCreateProperties()
{
    DUChainReadLocker lock (DUChain::lock());

    Declaration* decl = cursorItemDeclaration();
    if (!decl)
    {
        KMessageBox::information (0, i18n ("No Declaration."));
        return;
    }

    Declaration* classDecl = 0;
    Declaration * classMembDecl = 0;

    if (decl->internalContext() && decl->internalContext()->type() == DUContext::Class)
    {
        classDecl = decl;
    }
    else
    {
        DUContext *context = decl->context();
        if (!context)
        {
            KMessageBox::information (0, i18n ("No Context"));
            return;
        }

        if (context->type() != DUContext::Class)
        {
            KMessageBox::information (0, i18n ("Context is not a class or a member."));
            return;
        }

        classDecl = context->owner();
        classMembDecl = decl;
    }

    ClassDeclaration *klass = dynamic_cast<ClassDeclaration*> (classDecl);
    if (!klass)
    {
        kDebug() << "Failed to cast to class";
        return;
    }
    ClassMemberDeclaration * klassM = dynamic_cast<ClassMemberDeclaration *> (classMembDecl);

    createProperties (klass, klassM);
}

void ReaderWriterPlugin::createProperties (ClassDeclaration *klass, ClassMemberDeclaration *selected)
{
    static PropertiesDialog *d = (
    {
        PropertiesDialog *d = new PropertiesDialog ( (QWidget *) ICore::self()->uiController()->activeMainWindow());
        connect (d, SIGNAL (okClicked()), this, SLOT (slotWriteProperties()));
        d;
    });
    static PropertyModel *m = new PropertyModel;
    static CppPropertyCreator pc;

    if (d->isVisible())
        return;
    pc.parse (m, klass, selected);
    d->setCaption (i18n ("Class: ") + klass->identifier().toString());
    d->setModal (false);
    d->show (m);
}

void ReaderWriterPlugin::slotWriteProperties()
{
    static CppPropertyCreator pc;
    PropertiesDialog *d = dynamic_cast<PropertiesDialog *> (this->sender());
    if (!d)
    {
        kDebug() << "sender() not instance of PropertiesDialog";
        return;
    }
    pc.create (d->model());
}

ContextMenuExtension ReaderWriterPlugin::contextMenuExtension (Context* context)
{
    Q_UNUSED (context);

    m_action->setEnabled (false);
    ContextMenuExtension ext;

    DUChainReadLocker lock (DUChain::lock());

    Declaration * decl = cursorItemDeclaration();

    if (!decl)
        return ext;

    if (decl->kind() == Declaration::Instance && !decl->isFunctionDeclaration() &&
        decl->context() && decl->context()->type() == DUContext::Class)
    {
        ext.addAction (ContextMenuExtension::EditGroup, m_action);
        m_action->setEnabled (true);
        return ext;
    }

    if (decl->internalContext() && decl->internalContext()->type() == DUContext::Class)
    {
        ext.addAction (ContextMenuExtension::EditGroup, m_action);
        m_action->setEnabled (true);
        return ext;
    }

    return ext;
}

Declaration* ReaderWriterPlugin::cursorItemDeclaration()
{
    IDocument* doc = ICore::self()->documentController()->activeDocument();
    if (!doc)
        return 0;

    KTextEditor::Document* textDoc = doc->textDocument();
    if (!textDoc)
        return 0;

    KTextEditor::View* view = textDoc->activeView();
    if (!view)
        return 0;

    SimpleCursor cursor (view->cursorPosition());
    Declaration* definition = DUChainUtils::itemUnderCursor (doc->url(), cursor);
    return definition;
}

#include "readerwriterplugin.moc"

// kate: indent-mode cstyle; space-indent on; indent-width 4; replace-tabs on; 
